module L=List;;
module A=Array;;

type polynome = { coeffs: float array };;
type equation = { term: float array array; degre: int; nbterms: int; mat_inv: float array array ; color: float array };;
type vec = float array
type ray = { orig: vec; dir: vec }
type scene = Rien | Equation of equation | Group of equation * scene

let vec3 x y z c = [| x; y; z; c |]
let init f = Array.init 4 (fun i -> Array.init 4 (f i))
let translate r =
  init (fun i j -> if i = j then 1. else if j = 3 then -. r.(i) else 0.)
let identity = translate (vec3 0. 0. 0. 1.)

let ( *| ) s r = [| s *. r.(0) ; s *. r.(1) ; s *. r.(2) ; s *. r.(3) |]
let ( +| ) a b = [| a.(0) +. b.(0) ; a.(1) +. b.(1) ; a.(2) +. b.(2) ; a.(3) +. b.(3) |]
let ( -| ) a b = [| a.(0) -. b.(0) ; a.(1) -. b.(1) ; a.(2) -. b.(2) ; a.(3) -. b.(3) |]

let ( *|| ) s r = [| [| s *. r.(0).(0) ; s *. r.(0).(1) ; s *. r.(0).(2) ; s *. r.(0).(3) |] ;
                     [| s *. r.(1).(0) ; s *. r.(1).(1) ; s *. r.(1).(2) ; s *. r.(1).(3) |] ;
                     [| s *. r.(2).(0) ; s *. r.(2).(1) ; s *. r.(2).(2) ; s *. r.(2).(3) |] ;
                     [| s *. r.(3).(0) ; s *. r.(3).(1) ; s *. r.(3).(2) ; s *. r.(3).(3) |] |]

let ( +|| ) s r = [| [| s.(0).(0) +. r.(0).(0) ; s.(0).(1) +. r.(0).(1) ; s.(0).(2) +. r.(0).(2) ; s.(0).(3) +. r.(0).(3) |] ;
                     [| s.(1).(0) +. r.(1).(0) ; s.(1).(1) +. r.(1).(1) ; s.(1).(2) +. r.(1).(2) ; s.(1).(3) +. r.(1).(3) |] ;
                     [| s.(2).(0) +. r.(2).(0) ; s.(2).(1) +. r.(2).(1) ; s.(2).(2) +. r.(2).(2) ; s.(2).(3) +. r.(2).(3) |] ;
                     [| s.(3).(0) +. r.(3).(0) ; s.(3).(1) +. r.(3).(1) ; s.(3).(2) +. r.(3).(2) ; s.(3).(3) +. r.(3).(3) |] |]

let ( -|| ) s r = [| [| s.(0).(0) -. r.(0).(0) ; s.(0).(1) -. r.(0).(1) ; s.(0).(2) -. r.(0).(2) ; s.(0).(3) -. r.(0).(3) |] ;
                     [| s.(1).(0) -. r.(1).(0) ; s.(1).(1) -. r.(1).(1) ; s.(1).(2) -. r.(1).(2) ; s.(1).(3) -. r.(1).(3) |] ;
                     [| s.(2).(0) -. r.(2).(0) ; s.(2).(1) -. r.(2).(1) ; s.(2).(2) -. r.(2).(2) ; s.(2).(3) -. r.(2).(3) |] ;
                     [| s.(3).(0) -. r.(3).(0) ; s.(3).(1) -. r.(3).(1) ; s.(3).(2) -. r.(3).(2) ; s.(3).(3) -. r.(3).(3) |] |]

let dot a b = a.(0) *. b.(0) +. a.(1) *. b.(1) +. a.(2) *. b.(2) +. a.(3) *. b.(3)
let unitise r = (1. /. sqrt (dot r r)) *| r
let pi = 4. *. atan 1.

let foi= float_of_int ;;
let iof = int_of_float;;

let mat_mul mat_1 mat_2 =
  let x0 = mat_1.(0).(0) *. mat_2.(0) +. mat_1.(0).(1) *. mat_2.(1) +. mat_1.(0).(2) *. mat_2.(2) +. mat_1.(0).(3) *. mat_2.(3) in
  let y0 = mat_1.(1).(0) *. mat_2.(0) +. mat_1.(1).(1) *. mat_2.(1) +. mat_1.(1).(2) *. mat_2.(2) +. mat_1.(1).(3) *. mat_2.(3) in
  let z0 = mat_1.(2).(0) *. mat_2.(0) +. mat_1.(2).(1) *. mat_2.(1) +. mat_1.(2).(2) *. mat_2.(2) +. mat_1.(2).(3) *. mat_2.(3) in
  let c0 = mat_1.(3).(0) *. mat_2.(0) +. mat_1.(3).(1) *. mat_2.(1) +. mat_1.(3).(2) *. mat_2.(2) +. mat_1.(3).(3) *. mat_2.(3) in
  [| x0 ; y0 ; z0 ; c0 |]

let sum4 f = f 0 +. f 1 +. f 2 +. f 3

let mat_mul_2 a b = init (fun i j -> sum4 (fun k -> a.(i).(k) *. b.(k).(j)))

let rot_x t = let cos, sin = cos t, sin t in [|[|    1.;    0.;    0.; 0. |];
                                               [|    0.;   cos;   sin; 0. |];
                                               [|    0.; -.sin;   cos; 0. |];
                                               [|    0.;    0.;    0.; 1. |]|]

let rot_y t = let cos, sin = cos t, sin t in [|[|   cos;    0.;   sin; 0. |];
                                               [|    0.;    1.;    0.; 0. |];
                                               [| -.sin;    0.;   cos; 0. |];
                                               [|    0.;    0.;    0.; 1. |]|]
let rot_z t = let cos, sin = cos t, sin t in [|[|   cos;   sin;    0.; 0. |];
                                               [| -.sin;   cos;    0.; 0. |];
                                               [|    0.;    0.;    1.; 0. |];
                                               [|    0.;    0.;    0.; 1. |]|]

let scale a b c = [|[|   1. /. a;    0.;    0.; 0. |];
                  [|    0.;  1. /.   b;    0.; 0. |];
                  [|    0.;    0.;    1. /. c; 0. |];
                  [|    0.;    0.;    0.; 1. |]|]

let rec choices i n =
   if i=0 then 1
   else if i>n then 0
   else if i=1 then n
   else ((choices (i-1) n) * (n - i + 1)) / i ;;

let rec puiss x k = if (k<0) then puiss (1. /. x) (0 - k)
    else if k=0 then 1.
    else if k mod 2 = 0 then puiss (x *. x) (k/2)
        else x *. (puiss x (k-1));;

let cosinus a b =
  let norm_a = a.(0) *. a.(0) +. a.(1) *. a.(1) +. a.(2) *. a.(2) +. a.(3) *. a.(3) in
  let norm_b = b.(0) *. b.(0) +. b.(1) *. b.(1) +. b.(2) *. b.(2) +. b.(3) *. b.(3) in
  let ab = a.(0) *. b.(0) +. a.(1) *. b.(1) +. a.(2) *. b.(2) +. a.(3) *. b.(3) in
  ab /. sqrt (norm_a *. norm_b);;

let to_bernstein a_i =
   let n = Array.length a_i - 1 (* n est le degre *) in
   let b_i = Array.make (n+1) 0. in
   (* p(x)= sum a_i x^i *)
   (* x^k = sum_(i=0..n) C(k,i)/C(k,n) B(i,n)(x) *)
   for k=0 to n do
      for i=k to n do
         let c_k_i = float_of_int (choices k i) in
         let c_k_n = float_of_int (choices k n) in
         b_i.(i) <- b_i.(i) +. a_i.(k) *. c_k_i /. c_k_n
      done
   done;
   b_i;;

let milieu x0 x1 = (x0 +. x1)/. 2.;;

let deCasteljau coeffs =
   let n = Array.length coeffs - 1 (* n est le degre *) in
   let left= Array.make (n+1) 0. and right= Array.make (n+1) 0. in
   let layers = Array.make (n+1) coeffs in
   for level=1 to n do
      layers.(level) <- Array.make (n+1-level) 0.;
      for k=0 to n-level do
        layers.(level).(k) <-
         milieu layers.(level-1).(k) layers.(level-1).(k+1)
      done;
   done;
   for k=0 to n do left.(k) <- layers.(k).(0);
       right.(k) <- layers.(n-k).(k)
        done;
   left, right;;

let minmax tableau =
   let mini = ref tableau.(0) and maxi= ref tableau.(0) in
   for i=1 to Array.length tableau - 1 do
      mini := min !mini tableau.(i);
      maxi := max !maxi tableau.(i)
   done;
   !mini, !maxi;;

(* b est le tableau des coeffs, dans la base de Bernstein *)
(* epsilon est la précision souhaitée *)
let roots_0_1_bernstein epsilon b =
   let rec search x0 x1 coeffs known_roots =
      let xm = milieu x0 x1 in
      let (mi,ma)=minmax coeffs in
   (* Printf.printf "x0=%f x1=%f in [%f %f]\n" x0 x1 mi ma; *)
      if ma < 0. || mi > 0. then known_roots
      else if x1 -. x0 < epsilon then xm::known_roots
           else let left,right=deCasteljau coeffs in
           search x0 xm left (search xm x1 right known_roots)
   in
   let x0= 0. and x1=1. in
   let roots= search x0 x1 b [] in  roots;;

let roots_0_1_canonic epsilon a =
   roots_0_1_bernstein epsilon (to_bernstein a);;
let roots_1_infini_canonic epsilon a =
   let a' = Array.of_list (List.rev (Array.to_list a)) in
   let roots' = roots_0_1_canonic epsilon a' in
   List.rev (List.map (function x-> 1. /. x) roots');;

let roots_pos_canonic epsilon pol =
   roots_0_1_canonic epsilon pol
      @ roots_1_infini_canonic epsilon pol;;

let roots_neg_canonic epsilon pol =
   let pol'= Array.copy pol in
   for i=1 to Array.length pol - 1 do
           if i mod 2 = 1 then pol'.(i) <- 0. -. pol'.(i)
        done;
   let roots' = roots_pos_canonic epsilon pol' in
   List.map (function x-> 0. -. x ) roots' ;;

(* le polynome est dans la base canonique :
   polynome.(i)=coeff de x^i *)
let solve epsilon polynome =
   roots_neg_canonic epsilon polynome
      @ roots_pos_canonic epsilon polynome;;

let hit_ray_canonic ray equation =
  let pol = Array.make (equation.degre+1) 0. in
  for t=0 to equation.nbterms - 1 do
    let t_xdeg = iof equation.term.(t).(0) in
    let t_ydeg = iof equation.term.(t).(1) in
    let t_zdeg = iof equation.term.(t).(2) in
    let t_coeff = equation.term.(t).(3) in
    for i=0 to t_xdeg do
      for j=0 to t_ydeg do
        for k=0 to t_zdeg do
          let x0 = ray.orig.(0) in
          let y0 = ray.orig.(1) in
          let z0 = ray.orig.(2) in
          let a = ray.dir.(0) in
          let b = ray.dir.(1) in
          let c = ray.dir.(2) in
          pol.(i+j+k) <- pol.(i+j+k) +. t_coeff *. foi(choices i t_xdeg) *. foi(choices j t_ydeg) *. foi(choices k t_zdeg) *. puiss x0 (t_xdeg-i) *. puiss y0 (t_ydeg-j) *. puiss z0 (t_zdeg-k) *. puiss a i *. puiss b j *. puiss c k
        done
      done
    done
  done;
  solve 0.00001 pol;;

let gradient equation vec =
  let aa = ref 0. in
  let bb = ref 0. in
  let cc = ref 0. in
  for t=0 to equation.nbterms - 1 do
    let t_xdeg = iof equation.term.(t).(0) in
    let t_ydeg = iof equation.term.(t).(1) in
    let t_zdeg = iof equation.term.(t).(2) in
    let t_coeff = equation.term.(t).(3) in
    if t_xdeg >= 1 then
	  aa := !aa +. t_coeff *. foi(t_xdeg) *. puiss vec.(0) (t_xdeg-1) *. puiss vec.(1) t_ydeg *. puiss vec.(2) t_zdeg;
	if t_ydeg >= 1 then
	  bb := !bb +. t_coeff *. foi(t_ydeg) *. puiss vec.(0) (t_xdeg) *. puiss vec.(1) (t_ydeg-1) *. puiss vec.(2) t_zdeg;
	if t_zdeg >= 1 then
	  cc := !cc +. t_coeff *. foi(t_zdeg) *. puiss vec.(0) (t_xdeg) *. puiss vec.(1) t_ydeg *. puiss vec.(2) (t_zdeg-1);
  done;
  vec3 !aa !bb !cc (!aa *. vec.(0) +. !bb *. vec.(1) +. !cc *. vec.(2));;
  
let roots ray equation =
  let roots' = hit_ray_canonic ray equation in
  if L.length roots' <= 0 then
	infinity
  else
    L.hd roots'


let intersect ray scene =
  let rec aux ((lambda, _, _) as hit) = function
    Equation equation ->
        let ray' = {orig = mat_mul equation.mat_inv ray.orig ; dir = mat_mul equation.mat_inv ray.dir} in
        let lambda' = roots ray' equation in
        if lambda' >= lambda then hit else
          let pt_inter = (ray'.orig +| lambda' *| ray'.dir) in
          let normal = gradient equation pt_inter in
          (lambda', normal, equation.color)
    | Rien -> hit
    | Group (equation, scenes) ->
	let ray' = {orig = mat_mul equation.mat_inv ray.orig ; dir = mat_mul equation.mat_inv ray.dir} in
        let lambda' = roots ray' equation in
        if lambda' >= lambda then 
          aux hit scenes
        else
	  let pt_inter = (ray'.orig +| lambda' *| ray'.dir) in
          let normal = gradient equation pt_inter in
          aux (lambda', normal, equation.color) scenes in
  aux (infinity, vec3 0. 0. 0. 0., vec3 0. 0. 0. 0.) scene

let rec ray_trace light ray scene =
  let lambda, normal, color = intersect ray scene in
  if lambda = infinity then (0., color) else
    let g = cosinus normal light in
    let coeff = 1. -. g in
    if coeff < 0. then (0., color) else
    if coeff > 1. then (1., color) else
      (coeff, color)

let get_color tab =
  if(Array.length tab >= 2) then
    let r = ref 0. in
    let g = ref 0. in
    let b = ref 0. in
    let col = ref 0 in
    for i=0 to Array.length tab.(1) - 1  do
      let n = Array.length tab.(1).(i) - 1 in
      if n = 3 then
        if tab.(1).(i).(0) = 3. then (
          r := tab.(1).(i).(1);
          g := tab.(1).(i).(2);
          b := tab.(1).(i).(3);
          col := 1;)
    done;
    if(!col = 1) then
      [| !r ; !g ; !b |]
    else
      [| 1. ; 1. ; 1. |]
  else
    [| 1. ; 1. ; 1. |]

let get_trans tab =
  let transformation = ref identity in
  if(Array.length tab >= 2) then
    for i=0 to Array.length tab.(1) - 1  do
      let n = Array.length tab.(1).(i) - 1 in
      if n = 3 then
        if tab.(1).(i).(0) = 2. then (
          let t = translate (vec3 tab.(1).(i).(1) tab.(1).(i).(2) tab.(1).(i).(3) 1.) in 
          transformation := mat_mul_2 !transformation t;)
        else if tab.(1).(i).(0) = 1. then(
          if tab.(1).(i).(1) <> 0. then
            let r = rot_x (tab.(1).(i).(1)) in
            transformation := mat_mul_2 !transformation r;
          if tab.(1).(i).(2) <> 0. then
            let r = rot_y (tab.(1).(i).(2)) in
            transformation := mat_mul_2 !transformation r;
          if tab.(1).(i).(3) <> 0. then
            let r = rot_z (tab.(1).(i).(2)) in
            transformation := mat_mul_2 !transformation r;)
        else if tab.(1).(i).(0) = 4. then(
            let s = scale tab.(1).(i).(1) tab.(1).(i).(2) tab.(1).(i).(3) in
            transformation := mat_mul_2 !transformation s;)
    done;
  !transformation

let max_degree eq =
  let max = ref 0. in
  for i=0 to Array.length eq - 1 do
    for j=0 to Array.length eq.(i) - 2 do
    if !max < eq.(i).(j) then
      max := eq.(i).(j)
    done;
  done;
  !max

let main fich =
  let lexbuf = Lexing.from_channel (open_in fich) in
  let parselist = Parse.main Lex.token lexbuf in
  let lequ = ref [{term = [|[| 0.|] |] ; degre= 0 ; nbterms= 0; mat_inv = [|[| 0.|] |] ; color = [| 0. |]}] in
  for i=0 to Array.length parselist - 1 do
    if(i = 0) then
      lequ := [{term = parselist.(i).(0) ; degre= int_of_float (max_degree parselist.(i).(0)) ; nbterms= Array.length parselist.(i).(0) ; mat_inv = get_trans parselist.(i); color = get_color parselist.(i)}]
    else
      lequ := {term = parselist.(i).(0) ; degre= int_of_float (max_degree parselist.(i).(0)) ; nbterms= Array.length parselist.(i).(0) ; mat_inv = get_trans parselist.(i) ; color = get_color parselist.(i)} :: !lequ
  done;
  !lequ

(* ajoute une liste d'elements a un tas *)
let rec ajouter liste tas = match liste with
     [] -> tas
|    x::q -> Group (x, (ajouter q tas));;

let () =
  let fich = match Sys.argv with
    | [| _; l |] -> l
    | _ -> "equation.txt" in
  let equ = main fich in
  let scene = ajouter equ Rien in
  let n = 512 and light = vec3 50. 50. (10.) 0. in
  Printf.printf "P3\n%d %d\n255\n" n n;
  for y = n - 1 downto 0 do
    for x = 0 to n - 1 do
      let aux x = float (x - n / 2) /. float n *. sin( pi /. 4. ) in
      let d = vec3 (aux x) (aux y) 1. 0. in
      let (g, color) = ray_trace light {orig= vec3 0. 0. (-10.) 1.; dir=d} scene in
      let g = g *. 255. in
      let r = iof(color.(0) *. g) in
      let v = iof(color.(1) *. g) in
      let b = iof(color.(2) *. g) in
      Printf.printf "%d %d %d\n" r v b;
    done
  done